import asyncio

from glQiwiApi import QiwiWrapper, RequestError

PUBLIC_KEY = '48e7qUxn9T7RyYE1MVZswX1FRSbE6iyCj2gCRwwF3Dnh5XrasNTx3BGPiMsyXQFNKQhvukniQG8RTVhYm3iPpzd6T6fUwiBX4WcjnHgoqxERdngWnH6EYfc7uBsGKPq4MF23dW4nUoixGqkoHj1YhjM7JyGfvh1o6fUdCHfX2uY8cfMxUFDwj8qRQgwPF'
SECRET_KEY = 'eyJ2ZXJzaW9uIjoiUDJQIiwiZGF0YSI6eyJwYXlpbl9tZXJjaGFudF9zaXRlX3VpZCI6ImJuMXZmNy0wMCIsInVzZXJfaWQiOiIzODA5NjgzMTc0NTkiLCJzZWNyZXQiOiI1MWY2MDc1MzkzYzgwZWZiY2FiM2Q5ZTVhNThjNjQ1NmE3ZWY4NjkxNDJkZjI0NjczNWYzNzZmZjkwODQwM2U4In19'


async def p2p_usage():
    async with QiwiWrapper(secret_p2p=SECRET_KEY, public_p2p=PUBLIC_KEY) as w:
        # bill id будет сгенерирован как str(uuid.uuid4()), если не был передан
        bill = await w.create_p2p_bill(
            amount=1,
            comment='Im using glQiwiApi'
        )
        print(bill)
        # Так можно проверить статус на оплаченный
        status = (await w.check_p2p_bill_status(
            bill_id=bill.bill_id
        )) == 'PAID'
        # Или, начиная с версии 0.2.0
        status = await bill.check()
        print(status)
        # Это выдаст ошибку, так как не передан api_access_token и phone_number
        # Вы можете в любой момент переназначить токен или номер
        try:
            await w.get_bills(rows=50)
        except RequestError as ex:
            print(ex)
        # Переназначаем токены
        w.api_access_token = '7f8e514786b0cd326cf604223ec91861'
        w.phone_number = '+380968317459'
        print(await w.get_bills(rows=20))


asyncio.run(p2p_usage())



# Здесь представлен пример без контекста async with
import asyncio

from glQiwiApi import QiwiWrapper

# Создаем объект кошелька и обязательно передаем without_context = True, иначе будут проблемы с aiohttp.ClientSession
wallet = QiwiWrapper(
    api_access_token='7f8e514786b0cd326cf604223ec91861',
    phone_number="+380968317459",
    without_context=True
)


async def main():
    async with QiwiWrapper(secret_p2p='my_p2p') as w:
        w.public_p2p = 'my_public_p2p'
        bill = await w.create_p2p_bill(amount=1)
        # new version
        new_status = await bill.check()
        # old version
        old_status = (await w.check_p2p_bill_status(bill.bill_id)) == 'PAID'


asyncio.run(main())



import asyncio
import datetime

from glQiwiApi import QiwiWrapper, RequestError

TOKEN = "YOUR_API_ACCESS_TOKEN"
WALLET = "+NUMBER"
PUBLIC_KEY = 'YOUR_PUBLIC_P2P_TOKEN'
SECRET_KEY = 'YOUR_SECRET_P2P_TOKEN'


async def basic_usage():
    async with QiwiWrapper(api_access_token=TOKEN, phone_number=WALLET, public_p2p=PUBLIC_KEY,
                           secret_p2p=SECRET_KEY) as wallet:
        # Так вы можете получить информацию по транзакции, зная её айди и тип
        print(await wallet.transaction_info(
            transaction_type='OUT', transaction_id=21249852701
        ))
        # Таким образом вы можете получить статистику киви кошелька
        # РАЗНИЦА МЕЖДУ end_date и start_date ДОЛЖНА БЫТЬ МЕНЬШЕ 90 ДНЕЙ
        stats = await wallet.fetch_statistics(
            start_date=datetime.datetime.now() - datetime.timedelta(days=10),
            end_date=datetime.datetime.now()
        )
        print(stats.out[0].amount)
        # Полная информация об аккаунте
        info = await wallet.account_info()
        # Получаем айпи адресс, с которого был совершен последний вход
        print(info.auth_info.ip)
        # Переводим деньги на другой кошелек, при этом получая айди платежа
        payment_id = await wallet.to_wallet(
            trans_sum=999,
            to_number="some_number",
            comment="I love glQiwiApi"
        )
        print(payment_id)
        # handling api exceptions and get json representation
        try:
            await wallet.to_wallet(to_number="+WRONG_NUMBER", trans_sum=999)
        except RequestError as ex:
            print(ex.json())


asyncio.run(basic_usage())



from typing import Optional, Dict, Any


class NoUrlFound(Exception):
    """Данная ошибка возникает при неправильной авторизации yoomoney"""


class RequestProxyError(Exception):
    """Возникает, если были переданы неправильные параметры запроса"""


ProxyError = Exception()


class InvalidCardNumber(Exception):
    """Ошибка, при передаче номера карты в неправильном формате"""


class RequestAuthError(Exception):
    """
    Ошибка при неправильной аунтефикации POST or GET data

    """


class InvalidToken(Exception):
    """Ошибка, возникающая, если был передан неверный токен"""


class InvalidData(Exception):
    """Ошибка возникает, если были переданы или получены невалид данные при запросе"""


class RequestError(Exception):
    """Возникает при ошибках сервиса или неправильной передаче параметров"""

    def __init__(self, message: str, status_code: str, additional_info: Optional[str] = None,
                 json_info: Optional[Dict[str, Any]] = None, *args) -> None:
        super().__init__(*args)
        self.message = message
        self.status_code = status_code
        self.additional_info = additional_info
        self._json_info = json_info

    def __str__(self) -> str:
        return f"code={self.status_code} doc={self.message}, additional_info={self.additional_info}"""

    def __repr__(self) -> str:
        return self.__str__()

    def json(self) -> str:
        import json

        if isinstance(self._json_info, dict):
            return json.dumps(self._json_info, indent=2, ensure_ascii=False)
        return self.__str__()


__all__ = [
    'InvalidData', 'NoUrlFound', 'RequestAuthError', 'RequestProxyError',
    'ProxyError', 'InvalidCardNumber', 'InvalidToken', 'RequestError'
]



from typing import Optional, List

from pydantic import Field, BaseModel

from glQiwiApi.data_models.basics import Sum
from glQiwiApi.utils import custom_load


class PassInfo(BaseModel):
    last_pass_change: str = Field(alias="lastPassChange")
    next_pass_change: str = Field(alias="nextPassChange")
    password_used: bool = Field(alias="passwordUsed")


class MobilePinInfo(BaseModel):
    last_mobile_pin_change: Optional[str] = Field(alias="lastMobilePinChange")
    mobile_pin_used: bool = Field(alias="mobilePinUsed")
    next_mobile_pin_change: str = Field(alias="nextMobilePinChange")


class PinInfo(BaseModel):
    pin_used: bool = Field(alias="pinUsed")


class AuthInfo(BaseModel):
    ip: str
    bound_email: Optional[str] = Field(alias="boundEmail", const=None)
    last_login_date: Optional[str] = Field(alias="lastLoginDate", const=None)
    email_settings: Optional[dict] = Field(alias="emailSettings", const=None)
    mobile_pin_info: MobilePinInfo = Field(alias="mobilePinInfo")
    pass_info: PassInfo = Field(alias="passInfo")
    person_id: int = Field(alias="personId")
    pin_info: PinInfo = Field(alias="pinInfo")
    registration_date: str = Field(alias="registrationDate")


class SmsNotification(BaseModel):
    price: Sum
    enabled: bool
    active: bool
    end_date: Optional[str] = Field(alias="endDate", const=None)


class IdentificationInfo(BaseModel):
    bank_alias: str = Field(alias="bankAlias")
    identification_level: str = Field(alias="identificationLevel")
    passport_expired: bool = Field(alias="passportExpired")


class NickName(BaseModel):
    nickname: Optional[str] = None
    can_change: bool = Field(alias="canChange")
    can_use: bool = Field(alias="canUse")
    description: str = ""


class Feature(BaseModel):
    feature_id: int = Field(alias="featureId")
    feature_value: str = Field(alias="featureValue")
    start_date: str = Field(alias="startDate")
    end_date: str = Field(alias="endDate")


class ContractInfo(BaseModel):
    blocked: bool = False
    contract_id: int = Field(alias="contractId")
    creation_date: str = Field(alias="creationDate")
    identification_info: List[IdentificationInfo] = Field(alias="identificationInfo")
    sms_notification: SmsNotification = Field(alias="smsNotification")
    nickname: NickName
    features: Optional[List[Feature]] = None
    blocked: bool


class UserInfo(BaseModel):
    default_pay_currency: str = Field(alias="defaultPayCurrency")
    default_pay_source: Optional[int] = Field(alias="defaultPaySource", const=None)
    default_pay_account_alias: Optional[str] = Field(alias="defaultPayAccountAlias", const=None)
    email: Optional[str] = None
    first_transaction_id: int = Field(alias="firstTxnId")
    language: str
    operator: str
    phone_hash: str = Field(alias="phoneHash")
    promo_enabled: Optional[bool] = Field(alias="promoEnabled", const=None)


class QiwiAccountInfo(BaseModel):
    auth_info: Optional[AuthInfo] = Field(alias="authInfo", const=None)
    contract_info: Optional[ContractInfo] = Field(alias="contractInfo", const=None)
    user_info: Optional[UserInfo] = Field(alias="userInfo", const=None)

    class Config:
        json_loads = custom_load


__all__ = [
    'QiwiAccountInfo'
]



from glQiwiApi.exceptions import RequestError
from glQiwiApi.qiwi import QiwiWrapper
from glQiwiApi.yoo_money import YooMoneyAPI

__all__ = [
    'QiwiWrapper',
    'YooMoneyAPI',
    'RequestError'
]

__version__ = '0.1.8'



from typing import Dict, Optional, Any

import glQiwiApi
from glQiwiApi.api import HttpXParser
from glQiwiApi.exceptions import RequestError


class CustomParser(HttpXParser):

    def __init__(self, without_context: bool, messages: Dict[str, str]) -> None:
        super(CustomParser, self).__init__()
        self._without_context = without_context
        self.messages = messages

    async def _request(self, *args, **kwargs):
        response = await super(CustomParser, self)._request(*args, **kwargs)
        if response.status_code != 200:
            self.raise_exception(str(response.status_code), json_info=response.response_data)

        if self._without_context:
            await self.session.close()

        return response

    def raise_exception(self, status_code: str, json_info: Optional[Dict[str, Any]] = None) -> None:
        message = self.messages.get(str(status_code))
        raise RequestError(message, status_code, additional_info=f"{glQiwiApi.__version__} version api",
                           json_info=json_info)



from typing import Literal, Optional, Union

from pydantic import BaseModel, Field

from glQiwiApi.data_models.basics import Sum
from glQiwiApi.utils import custom_load


class Provider(BaseModel):
    id: int
    """ID провайдера в QIWI Wallet"""

    short_name: str = Field(alias="shortName")
    """краткое наименование провайдера"""

    long_name: Optional[str] = Field(alias="longName", const=None)
    """развернутое наименование провайдера"""

    logo_url: Optional[str] = Field(alias="logoUrl", const=None)
    """ссылка на логотип провайдера"""

    description: Optional[str] = None
    """описание провайдера (HTML)"""

    keys: Optional[Union[str, list]] = None
    """список ключевых слов"""

    site_url: Optional[str] = Field(alias="siteUrl", const=None)
    """сайт провайдера"""


class Transaction(BaseModel):
    transaction_id: int = Field(alias="txnId")
    """	ID транзакции в сервисе QIWI Кошелек"""

    person_id: int = Field(alias="personId")
    """Номер кошелька"""

    date: str
    """
    Для запросов истории платежей - Дата/время платежа, во временной зоне запроса (см. параметр startDate).
    Для запросов данных о транзакции - Дата/время платежа, время московское
    """

    error_code: Optional[int] = Field(alias="errorCode", const=None)
    """Код ошибки платежа"""

    error: Optional[str] = None
    """Описание ошибки"""

    status: Literal['WAITING', 'SUCCESS', 'ERROR']
    """
    Статус платежа. Возможные значения:
    WAITING - платеж проводится,
    SUCCESS - успешный платеж,
    ERROR - ошибка платежа.
    """

    type: Literal['IN', 'OUT', 'QIWI_CARD']
    """
    Тип платежа. Возможные значения:
    IN - пополнение,
    OUT - платеж,
    QIWI_CARD - платеж с карт QIWI (QVC, QVP).
    """

    status_text: str = Field(alias="statusText")
    """Текстовое описание статуса платежа"""

    trm_transaction_id: str = Field(alias="trmTxnId")
    """Клиентский ID транзакции"""

    to_account: str = Field(alias="account")
    """
    Для платежей - номер счета получателя.
    Для пополнений - номер отправителя, терминала или название агента пополнения кошелька
    """

    sum: Sum
    """Данные о сумме платежа или пополнения."""

    commission: Sum
    """Данные о комиссии"""

    total: Sum
    """Общие данные о платеже в формате объекта Sum"""

    provider: Provider
    """Данные о провайдере."""

    source: Provider
    """Служебная информация"""

    comment: Optional[str] = None
    """	Комментарий к платежу"""

    currency_rate: int = Field(alias="currencyRate")
    """Курс конвертации (если применяется в транзакции)"""

    class Config:
        json_loads = custom_load



import unittest

from glQiwiApi import YooMoneyAPI
from glQiwiApi.abstracts import AioTestCase
from glQiwiApi.data import Payment, AccountInfo, Operation

TOKEN = 'some_token'


class YooMoneyTest(AioTestCase):
    def setUp(self) -> None:
        self.w = YooMoneyAPI(TOKEN)

    async def test_build_url_to_get_token(self):
        url = await YooMoneyAPI.build_url_for_auth(
            scope=['account-info', "operation-history", 'operation-details'],
            client_id='some_client_id'
        )
        self.assertEqual(isinstance(url, str), True)

    async def test_get_access_token(self):
        token = await YooMoneyAPI.get_access_token(
            code='some_code',
            client_id='client_id'
        )
        self.assertEqual(isinstance(token, str), True)

    async def test_account_info(self) -> None:
        account_info = await self.w.account_info()
        self.assertEqual(isinstance(account_info, AccountInfo), True)

    async def test_transaction_history(self) -> None:
        transactions = await self.w.transactions()
        self.assertEqual(all(isinstance(transaction, Operation) for transaction in transactions), True)

    async def test_send(self):
        payment = await self.w.send(
            to_account='4100116633099701',
            protect=True,
            amount=2,
            comment_for_history='test_comment_for_history',
            comment='test_comment_for_receiver'
        )
        self.assertEqual(isinstance(payment, Payment), True)
        self.assertEqual(hasattr(payment, 'payment_id'), True)


if __name__ == '__main__':
    unittest.main()



import functools
import uuid
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Union, Optional, Dict, Literal, List

import aiofiles

from glQiwiApi.abstracts import AbstractPaymentWrapper
from glQiwiApi.aiohttp_custom_api import CustomParser
from glQiwiApi.data import Response, Identification
from glQiwiApi.exceptions import InvalidData, InvalidCardNumber
from glQiwiApi.data_models import QiwiAccountInfo, Transaction, Bill, BillError, Statistic, Limit, Account
from glQiwiApi.data_models.basics import Sum, Commission
from glQiwiApi.qiwi.basic_qiwi_config import *
from glQiwiApi.utils import datetime_to_str_in_iso, DataFormatter, multiply_objects_parse, simple_multiply_parse


class QiwiWrapper(AbstractPaymentWrapper):
    """
    Класс, реализующий обработку запросов к киви, используя основной класс HttpXParser,
    удобен он тем, что не просто отдает json подобные объекты, а всё это конвертирует в python датаклассы.
    Вы также можете импортировать и пользоваться данным классом для парсинга и собственных целей,
    по сути, это обвертка, и вы можете написать такие запросы для любой платежной системы или сайта
    """

    def __init__(self,
                 api_access_token: Optional[str] = None,
                 phone_number: Optional[str] = None,
                 secret_p2p: Optional[str] = None,
                 public_p2p: Optional[str] = None,
                 without_context: bool = False) -> None:
        """
        :param api_access_token: токен, полученный с https://qiwi.com/api
        :param phone_number: номер вашего телефона с +
        :param secret_p2p: секретный ключ, полученный с https://p2p.qiwi.com/
        :param public_p2p: публичный ключ, полученный с https://p2p.qiwi.com/
        :param without_context: булевая переменная, указывающая будет ли объект класса "глобальной" переменной
         или будет использована в async with контексте
        """
        self._parser = CustomParser(without_context=without_context, messages=ERROR_CODE_NUMBERS)
        self.api_access_token = api_access_token
        self.phone_number = phone_number
        self._formatter = DataFormatter()
        self.public_p2p = public_p2p
        self.secret_p2p = secret_p2p

    async def __aenter__(self) -> 'QiwiWrapper':
        """Создаем сессию, чтобы не пересоздавать её много раз"""
        self._parser.create_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Закрываем сессию при выходе"""
        await self._parser.session.close()

    def _auth_token(self, headers: dict, p2p: bool = False) -> dict:
        headers['Authorization'] = headers['Authorization'].format(
            token=self.api_access_token if not p2p else self.secret_p2p
        )
        return headers

    async def to_card(
            self,
            trans_sum: Union[float, int],
            to_card: str
    ) -> Union[Response, str]:
        """
        Метод для отправки средств на карту.
        Более подробная документация https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards

        :param trans_sum: сумма перевода
        :param to_card: номер карты получателя
        :return
        """
        data = deepcopy(QIWI_TO_CARD)
        data.json['sum']['amount'] = trans_sum
        data.json['fields']['account'] = to_card
        data.headers = self._auth_token(headers=data.headers)
        privat_card_id = await self._detect_card_number(card_number=to_card)
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/sinap/api/v2/terms/' + privat_card_id + '/payments',
                headers=data.headers,
                json=data.json,
                get_json=True
        ):
            try:
                return response.response_data.get('id')
            except (KeyError, TypeError):
                return response

    async def _detect_card_number(self, card_number: str) -> str:
        """
        Метод для получения индентификатора карты

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(DEFAULT_QIWI_HEADERS)
        headers.update(
            {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        async for response in self._parser.fetch(
                url='https://qiwi.com/card/detect.action',
                headers=headers,
                method='POST',
                data={
                    'cardNumber': card_number
                }
        ):
            try:
                return response.response_data.get('message')
            except KeyError:
                raise InvalidCardNumber('Invalid card number or qiwi_models api is not response') from None

    async def _detect_mobile_number(self, phone_number: str):
        """
        Метод для получения индентификатора телефона

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(DEFAULT_QIWI_HEADERS)
        headers.update(
            {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        async for response in self._parser.fetch(
                url='https://qiwi.com/mobile/detect.action',
                headers=headers,
                method='POST',
                data={
                    'phone': phone_number
                }
        ):
            try:
                return response.response_data.get('message')
            except KeyError:
                raise InvalidCardNumber('Invalid card number or qiwi_models api is not response') from None

    async def get_balance(self) -> Sum:
        """Метод для получения баланса киви"""
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/funding-sources/v2/persons/' + self.phone_number + '/accounts',
                headers=headers,
                method='GET',
                get_json=True
        ):
            return Sum.parse_raw(response.response_data['accounts'][0]['balance'])

    async def transactions(
            self,
            rows_num: int = 50,
            operation: Literal['ALL', 'IN', 'OUT', 'QIWI_CARD', 'ALL'] = 'ALL',
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None
    ) -> Union[Optional[List[Transaction]], dict]:
        """
        Метод для получения транзакций на счёту
        Более подробная документация https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param rows_num: кол-во транзакций, которые вы хотите получить
        :param operation: Тип операций в отчете, для отбора.
        :param start_date:Начальная дата поиска платежей. Используется только вместе с end_date.
        :param end_date: онечная дата поиска платежей. Используется только вместе со start_date.

        """
        if rows_num > 50:
            raise InvalidData('Можно проверять не более 50 транзакций')

        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        payload_data = {
            'rows': rows_num,
            'operation': operation
        }

        if isinstance(start_date, (datetime, timedelta)) and isinstance(end_date, (datetime, timedelta)):
            if (end_date - start_date).total_seconds() > 0:
                payload_data.update(
                    {
                        'startDate': datetime_to_str_in_iso(start_date),
                        'endDate': datetime_to_str_in_iso(end_date)
                    }
                )
            else:
                raise InvalidData('end_date не может быть больше чем start_date')
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/payment-history/v2/persons/' + self.phone_number.replace('+', '') + '/payments',
                params=payload_data,
                headers=headers,
                method='GET',
                get_json=True
        ):
            return multiply_objects_parse(
                lst_of_objects=(response.response_data.get('data'),),
                model=Transaction
            )

    async def to_wallet(
            self,
            to_number: str,
            trans_sum: Union[str, float, int],
            currency: str = '643',
            comment: str = '+comment+') -> Union[str, Response]:
        """
        Метод для перевода денег на другой кошелек\n
        Подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_number: номер получателя
        :param trans_sum: кол-во денег, которое вы хотите перевести
        :param currency: особенный код валюты
        :param comment: комментарий к платёжу
        """
        data = self._formatter.set_data_to_wallet(
            data=deepcopy(QIWI_TO_WALLET),
            to_number=to_number,
            trans_sum=trans_sum,
            currency=currency,
            comment=comment
        )
        data.headers = self._auth_token(
            headers=data.headers
        )
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/sinap/api/v2/terms/99/payments',
                json=data.json,
                headers=data.headers
        ):
            try:
                return response.response_data['transaction']['id']
            except (KeyError, TypeError):
                return response

    async def transaction_info(
            self,
            transaction_id: Union[str, int],
            transaction_type: Literal['IN', 'OUT']) \
            -> Optional[Transaction]:
        """
        Метод для получения полной информации о транзакции\n
        Подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#txn_info

        :param transaction_id: номер транзакции
        :param transaction_type: тип транзакции, может быть только IN или OUT
        :return: Transaction object
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        payload_data = {
            'type': transaction_type
        }
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/payment-history/v1/transactions/' + str(transaction_id),
                headers=headers,
                params=payload_data,
                method='GET',
                get_json=True
        ):
            return Transaction.parse_raw(response.response_data)

    async def check_restriction(self) -> Union[List[Dict[str, str]], Exception]:
        """
        Метод для проверки ограничений на вашем киви кошельке\n
        Подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#restrictions

        :return: Список, где находиться словарь с ограничениями, если ограничений нет - возвращает пустой список
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/person-profile/v1/persons/' + self.phone_number + '/status/restrictions',
                headers=headers,
                method='GET'
        ):
            return response.response_data

    async def get_identification(self) -> List[Identification]:
        """
        Функция, которая позволяет получить данные индентификации вашего киви кошелька
        Более подробная документация https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#ident

        :return: Response object
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/identification/v1/persons/' + self.phone_number + '/identification',
                method='GET',
                headers=headers
        ):
            return self._formatter.format_objects(
                iterable_obj=(response.response_data,),
                obj=Identification,
                transfers=IDENTIFICATION_TRANSFER
            )

    async def check_transaction(
            self,
            amount: Union[int, float],
            transaction_type: Literal['IN', 'OUT', 'QIWI_CARD'] = 'IN',
            sender_number: Optional[str] = None,
            rows_num: int = 50,
            comment: Optional[str] = None) -> bool:
        """
        Метод для проверки транзакции.\n Рекомендуется использовать только если вы не можете написать свой обработчик.\n
        Данный метод использует self.transactions(rows_num=rows_num) для получения платежей.\n
        Для небольшой оптимизации вы можете уменьшить rows_num задав его, однако это не гарантирует правильный результат

        :param amount: сумма платежа
        :param transaction_type: тип платежа
        :param sender_number: номер получателя
        :param rows_num: кол-во платежей, которое будет проверяться
        :param comment: комментарий, по которому будет проверяться транзакция
        :return: bool, есть ли такая транзакция в истории платежей
        """
        if transaction_type not in ['IN', 'OUT', 'QIWI_CARD']:
            raise InvalidData('Вы ввели неправильный метод транзакции')
        elif rows_num > 50:
            raise InvalidData('Можно проверять не более 50 транзакций')
        transactions = await self.transactions(rows_num=rows_num)
        for transaction in transactions:
            if float(transaction.sum.amount) >= amount and transaction.type == transaction_type:
                if transaction.comment == comment and transaction.to_account == sender_number:
                    return True
                elif comment and sender_number:
                    continue
                elif transaction.to_account == sender_number:
                    return True
                elif sender_number:
                    continue
                elif transaction.comment == comment:
                    return True
        return False

    async def get_limits(self) -> Dict[str, Limit]:
        """
        Функция для получения лимитов по счёту киви кошелька\n
        Возвращает лимиты по кошельку в виде списка, если лимита по определенной стране нет, то не включает его в список
        Подробная документация https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits

        :return: Limit object of Limit(pydantic)
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))

        payload = {}

        for index, limit_type in enumerate(LIMIT_TYPES):
            payload['types[' + str(index) + ']'] = limit_type

        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/qw-limits/v1/persons/' + self.phone_number.replace("+", "") + '/actual-limits',
                get_json=True,
                headers=headers,
                params=payload,
                method='GET'
        ):
            limits = {}
            for code, limit in response.response_data.get("limits").items():
                if not limit:
                    continue
                limits.update({code: Limit.parse_raw(limit[0])})
            return limits

    async def get_list_of_cards(self) -> dict:
        """
        Данный метод позволяет вам получить список ваших карт. Пока ещё в разработке, стабильность сомнительна

        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/cards/v1/cards?vas-alias=qvc-master',
                method='GET',
                headers=headers
        ):
            return response.response_data

    async def authenticate(
            self,
            birth_date: str,
            first_name: str,
            last_name: str,
            patronymic: str,
            passport: str,
            oms: str = "",
            inn: str = "",
            snils: str = ""
    ) -> Optional[Dict[str, bool]]:
        """
        Данный запрос позволяет отправить данные для идентификации вашего QIWI кошелька.
        Допускается идентифицировать не более 5 кошельков на одного владельца

        Для идентификации кошелька вы обязательно должны отправить ФИО, серию/номер паспорта и дату рождения.\n
        Если данные прошли проверку, то в ответе будет отображен
        ваш ИНН и упрощенная идентификация кошелька будет установлена.
        В случае если данные не прошли проверку, кошелек остается в статусе "Минимальный".
        :param birth_date: Дата рождения в виде строки формата 1998-02-11
        :param first_name: Ваше имя
        :param last_name: Ваша фамилия
        :param patronymic: Ваше отчество
        :param passport: Серия / Номер паспорта. Пример 4400111222
        :param oms:
        :param snils:
        :param inn:
        """

        payload = {
            "birthDate": birth_date,
            "firstName": first_name,
            "inn": inn,
            "lastName": last_name,
            "middleName": patronymic,
            "oms": oms,
            "passport": passport,
            "snils": snils
        }
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/identification/v1/persons/' + self.phone_number.replace("+",
                                                                                              "") + "/identification",
                data=payload,
                headers=headers
        ):
            if response.ok and response.status_code == 200:
                return {'success': True}

    async def get_bills(self, rows: int) -> List[Bill]:
        """
        Метод получения списка неоплаченных счетов вашего кошелька.
        Список строится в обратном хронологическом порядке.
        По умолчанию, список разбивается на страницы по 50 элементов в каждой, но вы можете задать другое количество
        элементов (не более 50). В запросе можно использовать фильтры по времени выставления счета,
        начальному идентификатору счета.
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        if rows > 50:
            raise InvalidData('Можно получить не более 50 счетов')
        params = {
            'rows': rows
        }
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/checkout-api/api/bill/search?statuses=READY_FOR_PAY',
                headers=headers,
                method='GET',
                params=params
        ):
            return simple_multiply_parse(response.response_data.get("bills"), Bill)

    async def create_p2p_bill(
            self,
            amount: int,
            bill_id: Optional[str] = None,
            comment: Optional[str] = None,
            life_time: Optional[datetime] = None,
            theme_code: Optional[str] = None,
            pay_source_filter: Optional[Literal['qw', 'card', 'mobile']] = None
    ) -> Union[Bill, BillError]:
        """
        Метод для выставление p2p счёта.
        Надежный способ для интеграции. Параметры передаются server2server с использованием авторизации.


        :param amount: сумма платежа
        :param bill_id: уникальный номер транзакции, если не передан, генерируется автоматически,
        :param life_time: дата, до которой счет будет доступен для оплаты.
        :param comment: комментарий к платёжу
        :param theme_code: специальный код темы
        :param pay_source_filter: При открытии формы будут отображаться только указанные способы перевода
        """
        if not self.public_p2p or not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')

        if not bill_id:
            bill_id = str(uuid.uuid4())

        life_time = datetime_to_str_in_iso((datetime.now() + timedelta(days=2)) if not life_time else life_time)
        data = deepcopy(P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        payload = self._formatter.set_data_p2p_create(
            wrapped_data=data,
            amount=amount,
            comment=comment,
            theme_code=theme_code,
            pay_source_filter=pay_source_filter,
            life_time=life_time
        )
        async for response in self._parser.fast().fetch(
                url=f'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}',
                json=payload,
                headers=headers,
                method='PUT'
        ):
            # unparsed_data = str(json.dumps(response.response_data))
            return Bill.parse_raw(response.response_data).initialize(self)

    async def check_p2p_bill_status(self, bill_id: str) -> Literal['WAITING', 'PAID', 'REJECTED', 'EXPIRED']:
        """
        Метод для проверки статуса p2p транзакции.\n
        Возможные типы транзакции: \n
        WAITING	Счет выставлен, ожидает оплаты	\n
        PAID	Счет оплачен	\n
        REJECTED	Счет отклонен\n
        EXPIRED	Время жизни счета истекло. Счет не оплачен\n
        Более подробная документация https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id: номер p2p транзакции
        :return: статус транзакции строкой
        """
        if not self.public_p2p or not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')
        data = deepcopy(P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        async for response in self._parser.fast().fetch(
                url=f'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}',
                method='GET',
                headers=headers
        ):
            return Bill.parse_raw(response.response_data).status.value

    async def reject_p2p_bill(self, bill_id: str) -> Bill:
        """
        Метод для отмены транзакции.

        :param bill_id: номер p2p транзакции
        :return: транзакцию в датаклассе
        """
        if not self.public_p2p or not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')
        data = deepcopy(P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        async for response in self._parser.fast().fetch(
                url=f'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}/reject',
                method='POST',
                headers=headers
        ):
            return Bill.parse_raw(response.response_data)

    @functools.lru_cache
    async def get_receipt(
            self,
            transaction_id: Union[str, int],
            transaction_type: Literal['IN', 'OUT', 'QIWI_CARD'],
            file_path: Optional[str] = None
    ) -> Union[bytearray, int]:
        """
        Метод для получения чека в формате байтов или файлом.

        :param transaction_id: айди транзакции, можно получить при вызове методе to_wallet, to_card
        :param transaction_type: тип транзакции может быть 'IN', 'OUT', 'QIWI_CARD'
        :param file_path: путь к файлу, куда вы хотите сохранить чек, если не указан, возвращает байты
        :return: pdf файл в байтовом виде или номер записанных байтов
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        data = {
            'type': transaction_type,
            'format': 'PDF'
        }
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/payment-history/v1/transactions/' + str(transaction_id) + '/cheque/file',
                headers=headers,
                method='GET',
                params=data
        ):
            if not file_path:
                return response.response_data

            async with aiofiles.open(file_path + '.pdf', 'wb') as file:
                return await file.write(response.response_data)

    async def commission(self, to_account: str, pay_sum: Union[int, float]) -> Commission:
        """
        Возвращается полная комиссия QIWI Кошелька за платеж в пользу указанного провайдера
        с учетом всех тарифов по заданному набору платежных реквизитов.

        :param to_account: номер карты или киви кошелька
        :param pay_sum: сумма, за которую вы хотите узнать коммиссию
        :return: Commission object
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        json_payload = deepcopy(ONLINE_COMMISSION_DATA)
        json_payload['purchaseTotals']['total']['amount'] = pay_sum
        json_payload['account'] = to_account.replace('+', '')
        special_code = "99" if len(to_account.replace('+', '')) <= 15 else (
            await self._detect_card_number(card_number=to_account))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/sinap/providers/' + special_code + '/onlineCommission',
                headers=headers,
                json=json_payload
        ):
            return Commission.parse_raw(response.response_data)

    async def account_info(self) -> QiwiAccountInfo:
        """Метод для получения информации об аккаунте"""
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/person-profile/v1/profile/current',
                headers=headers,
                method='GET',
                get_json=True
        ):
            return QiwiAccountInfo.parse_raw(response.response_data)

    async def fetch_statistics(
            self, start_date: Union[datetime, timedelta], end_date: Union[datetime, timedelta],
            operation: Literal['ALL', 'IN', 'OUT', 'QIWI_CARD'] = 'ALL',
            sources: Optional[List[Literal['QW_RUB', 'CARD', 'QW_USD', 'QW_EUR', 'MK']]] = None
    ) -> Statistic:
        """
        Данный запрос используется для получения сводной статистики по суммам платежей за заданный период.\n
        Более подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param start_date: Начальная дата периода статистики. Обязательный параметр
        :param end_date: Конечная дата периода статистики. Обязательный параметр
        :param operation: Тип операций, учитываемых при подсчете статистики. Допустимые значения:
            ALL - все операции,
            IN - только пополнения,
            OUT - только платежи,
            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
            По умолчанию ALL.
        :param sources: Источники платежа, по которым вернутся данные.
            QW_RUB - рублевый счет кошелька,
            QW_USD - счет кошелька в долларах,
            QW_EUR - счет кошелька в евро,
            CARD - привязанные и непривязанные к кошельку банковские карты,
            MK - счет мобильного оператора. Если не указан, учитываются все источники платежа.
        """
        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))

        if isinstance(start_date, (datetime, timedelta)) and isinstance(end_date, (datetime, timedelta)):
            delta = end_date - start_date
            if delta.days > 90:
                raise InvalidData('Максимальный период для выгрузки статистики - 90 календарных дней.')
        else:
            raise InvalidData('Вы передали значения начальной и конечной даты в неправильном формате.')

        params = {
            'startDate': datetime_to_str_in_iso(start_date),
            'endDate': datetime_to_str_in_iso(end_date),
            'operation': operation,
        }

        if sources:
            params.update({'sources': ' '.join(sources)})
        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/payment-history/v2/persons/' + self.phone_number.replace("+",
                                                                                               "") + '/payments/total',
                params=params,
                headers=headers,
                get_json=True,
                method='GET'
        ):
            return Statistic.parse_raw(response.response_data)

    async def list_of_balances(self) -> List[Account]:
        """
        Запрос выгружает текущие балансы счетов вашего QIWI Кошелька.
        Более подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list
        """

        async for response in self._parser.fast().fetch(
                url=BASE_QIWI_URL + '/funding-sources/v2/persons/' + self.phone_number.replace("+", "") + '/accounts',
                headers=self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS)),
                method='GET'
        ):
            return simple_multiply_parse(response.response_data.get('accounts'), Account)



from typing import Literal, Optional

from pydantic import BaseModel, Field

from glQiwiApi.data_models.basics import OptionalSum, BillMixin
from glQiwiApi.utils import custom_load


class Customer(BaseModel):
    phone: str
    email: str
    account: str


class BillStatus(BaseModel):
    value: Literal['WAITING', 'PAID', 'REJECTED', 'EXPIRED']
    changed_datetime: str = Field(alias="changedDateTime")


class CustomFields(BaseModel):
    pay_sources_filter: str = Field(alias="paySourcesFilter")
    theme_code: str = Field(alias="themeCode")


class BillError(BaseModel):
    service_name: str = Field(alias="serviceName")
    error_code: str = Field(alias="errorCode")
    description: str
    user_message: str = Field(alias="userMessage")
    datetime: str = Field(alias="dateTime")
    trace_id: str = Field(alias="traceId")

    class Config:
        json_loads = custom_load


class Bill(BaseModel, BillMixin):
    site_id: str = Field(alias="siteId")
    bill_id: str = Field(alias="billId")
    amount: OptionalSum
    status: BillStatus
    creation_date_time: str = Field(alias="creationDateTime")
    expiration_date_time: str = Field(alias="expirationDateTime")
    pay_url: str = Field(alias="payUrl")
    custom_fields: Optional[CustomFields] = Field(alias="customFields", const=None)
    customer: Optional[Customer] = None

    class Config:
        extra = 'allow'
        json_loads = custom_load


__all__ = [
    'Bill', 'BillError'
]



from typing import Union, Any, Callable, TypeVar

from glQiwiApi.data import Identification, AccountInfo, Operation, \
    OperationDetails, PreProcessPaymentResponse, Payment, IncomingTransaction, WrapperData
from glQiwiApi.data_models import Transaction, QiwiAccountInfo, Account, Limit, Bill, BillError, Statistic

BasicTypes = Union[
    Transaction, Identification, AccountInfo, Operation, OperationDetails,
    PreProcessPaymentResponse, Payment, IncomingTransaction
]

PydanticTypes = Union[
    Transaction, QiwiAccountInfo, Account, Limit, Bill, BillError, Statistic
]

F = TypeVar('F', bound=Callable[..., Any])

__all__ = ('BasicTypes', 'F', 'WrapperData', 'PydanticTypes')



import abc
import asyncio
import unittest
from typing import AsyncGenerator


class AbstractPaymentWrapper(abc.ABC):
    @abc.abstractmethod
    async def transactions(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def transaction_info(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def get_balance(self) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def account_info(self, *args, **kwargs) -> None:
        raise NotImplementedError()


class AioTestCase(unittest.TestCase):

    # noinspection PyPep8Naming
    def __init__(self, methodName='runTest', loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self._function_cache = {}
        super(AioTestCase, self).__init__(methodName=methodName)

    def coroutine_function_decorator(self, func):
        def wrapper(*args, **kw):
            return self.loop.run_until_complete(func(*args, **kw))

        return wrapper

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if asyncio.iscoroutinefunction(attr):
            if item not in self._function_cache:
                self._function_cache[item] = self.coroutine_function_decorator(attr)
            return self._function_cache[item]
        return attr


class AbstractParser(abc.ABC):

    @abc.abstractmethod
    async def _request(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def fetch(self, *args, **kwargs) -> AsyncGenerator:
        raise NotImplementedError()

    async def raise_exception(self, *args, **kwargs) -> None:
        """Метод для кастомной обработки исключений и лучшего логирования"""



import functools as ft
import json
import re
import time
from dataclasses import is_dataclass
from datetime import datetime
from typing import Optional, Union, Type

import pytz
from pydantic import ValidationError
from pytz.reference import LocalTimezone

Local = LocalTimezone()


def measure_time(func):
    @ft.wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.monotonic()
        await func(*args, **kwargs)
        print(f'Executed for {time.monotonic() - start_time} secs')

    return wrapper


def datetime_to_str_in_iso(obj, yoo_money_format=False):
    if not isinstance(obj, datetime):
        return ''
    if yoo_money_format:
        # Приводим время к UTC, так как юмани апи принимает именно в таком формате
        return pytz.utc.localize(obj).replace(tzinfo=None).isoformat(' ').replace(" ", "T") + "Z"
    return obj.isoformat(' ').replace(" ", "T") + re.findall(r'[+]\d{2}[:]\d{2}', str(datetime.now(tz=Local)))[0]


def parse_auth_link(response_data):
    return re.findall(r'https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+', str(response_data))[0]


def parse_headers(content_json=False, auth=False):
    """
    Функция для добавления некоторых заголовков в запрос

    """
    headers = {
        'Host': 'yoomoney.ru',
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    if content_json:
        headers.update(
            {'Accept': 'application/json'}
        )
    if auth:
        headers.update(
            {'Authorization': 'Bearer {token}'}
        )
    return headers


def format_objects_for_fill(data, transfers):
    for key, value in data.copy().items():
        if hasattr(transfers, 'get'):
            if key in transfers.keys():
                data.update({transfers.get(key): value})
                data.pop(key)
    return data


class DataFormatter:

    @staticmethod
    def set_data_to_wallet(data, to_number, trans_sum, comment, currency):
        data.json['sum']['amount'] = str(trans_sum)
        data.json['sum']['currency'] = currency
        data.json['fields']['account'] = to_number
        data.json['comment'] = comment
        data.headers.update({'User-Agent': 'Android v3.2.0 MKT'})
        return data

    def format_objects(self, iterable_obj, obj, transfers=None):
        """Метод для форматирования объектов, которые приходят от апи, pydantic usage soon..."""
        kwargs = {}
        objects = []
        for transaction in iterable_obj:
            for key, value in transaction.items():
                if key in obj.__dict__.get('__annotations__').keys() or key in transfers.keys():
                    try:
                        fill_key = key if not transfers.get(key) else transfers.get(key)
                    except AttributeError:
                        fill_key = key
                    sp_obj: Optional[Union[str, Type]] = obj.__dict__.get('__annotations__').get(fill_key)
                    if is_dataclass(sp_obj):
                        try:
                            kwargs.update({fill_key: sp_obj(**value)})
                        except TypeError:
                            kwargs.update({fill_key: sp_obj(**format_objects_for_fill(value, transfers))})
                        continue
                    kwargs.update({fill_key: value})
            objects.append(obj(**kwargs))
            kwargs = {}
        return objects

    def set_data_p2p_create(self, wrapped_data, amount, life_time, comment, theme_code, pay_source_filter):
        wrapped_data.json['amount']['value'] = str(amount)
        wrapped_data.json['comment'] = comment
        wrapped_data.json['expirationDateTime'] = life_time
        if pay_source_filter in ['qw', 'card', 'mobile']:
            wrapped_data.json['customFields']['paySourcesFilter'] = pay_source_filter
        if isinstance(theme_code, str):
            wrapped_data.json['customFields']['theme_code'] = theme_code
        if not isinstance(theme_code, str) and pay_source_filter not in ['qw', 'card', 'mobile']:
            wrapped_data.json.pop('customFields')
        return wrapped_data.json


#
def multiply_objects_parse(lst_of_objects, model):
    objects = []
    for obj in lst_of_objects:
        try:
            if isinstance(obj, dict):
                obj: str
                objects.append(model.parse_raw(obj))
                continue
            for detached_obj in obj:
                objects.append(model.parse_raw(detached_obj))
        except ValidationError as ex:
            print(ex.json(indent=4))
    return objects


def simple_multiply_parse(lst_of_objects, model):
    objects = []
    for obj in lst_of_objects:
        objects.append(model.parse_raw(obj))
    return objects


def dump_response(func):
    """Декоратор, который гарантирует получения json данных в запросе"""

    @ft.wraps(func)
    async def wrapper(*args, **kwargs):
        kwargs.update({'get_json': True})
        return await func(*args, **kwargs)

    return wrapper


def custom_load(data):
    return json.loads(json.dumps(data))



from dataclasses import dataclass
from enum import Enum
from http.cookies import SimpleCookie
from typing import Literal, Optional, Union, Dict, List, Any

from aiohttp.typedefs import RawHeaders
from aiosocksy import Socks5Auth, Socks4Auth

from glQiwiApi.exceptions import ProxyError


@dataclass
class Response:
    status_code: int
    response_data: Optional[Union[dict, str, bytes, bytearray, Exception]] = None
    url: Optional[str] = None
    raw_headers: Optional[RawHeaders] = None
    cookies: Optional[SimpleCookie] = None
    ok: bool = False
    content_type: Optional[str] = None
    host: Optional[str] = None

    @classmethod
    def bad_response(cls) -> 'Response':
        return cls(
            status_code=500,
            response_data=ProxyError
        )


@dataclass
class ProxyService:
    login: str
    password: str
    ip_address: str
    service_type: Literal['SOCKS5', 'SOCKS4'] = 'SOCKS5'
    proxy_auth: Optional[Socks5Auth] = None
    socks_url: Optional[str] = None

    def get_proxy(self) -> Dict[str, Union[str, Socks5Auth]]:
        if not isinstance(self.proxy_auth, (Socks5Auth, Socks4Auth)):
            self.proxy_auth = Socks5Auth(
                login=self.login,
                password=self.password
            )

        self.socks_url = '{socks_type}://{ip_address}'.format(
            socks_type=self.service_type.lower(),
            ip_address=self.ip_address
        )
        return dict(
            proxy_auth=self.proxy_auth,
            proxy=self.socks_url
        )


proxy_list = (
    ProxyService(
        login='6TA3h0',
        password='3qHCjh',
        ip_address='91.241.47.240:8000'
    ),
)


@dataclass
class WrapperData:
    headers: Dict[str, Union[str, int]]
    data: Dict[str, Union[str, Dict[str, str]]] = None
    json: Dict[str, Union[str, Dict[str, str]]] = None
    cookies: Optional[Dict[str, Union[str, int]]] = None


@dataclass(frozen=True)
class Identification:
    identification_id: int
    first_name: str
    middle_name: str
    last_name: str
    birth_date: str
    passport: str
    inn: str
    snils: str
    oms: str
    type: str

# YooMoney objects

@dataclass
class BalanceDetails:
    total: float
    available: float
    deposition_pending: Optional[float] = None
    blocked: Optional[float] = None
    debt: Optional[float] = None
    hold: Optional[float] = None


@dataclass
class AccountInfo:
    account: str
    """Номер счета"""

    balance: float
    """Баланс счета"""

    currency: str
    """Код валюты счета пользователя. Всегда 643 (рубль РФ по стандарту ISO 4217)."""

    identified: bool
    """Индентификацирован ли кошелек"""

    account_type: str
    """
    Тип счета пользователя. Возможные значения:
    personal — счет пользователя в ЮMoney;
    professional — профессиональный счет в ЮMoney
    """

    account_status: str
    """
    Статус пользователя. Возможные значения:
    anonymous — анонимный счет;
    named — именной счет;
    identified — идентифицированный счет.
    """

    balance_details: BalanceDetails
    """
    Расширенная информация о балансе. 
    По умолчанию этот блок отсутствует.
    Блок появляется, если сейчас или когда-либо ранее были зачисления в очереди, задолженности, блокировки средств.
    """

    cards_linked: Optional[List[Dict[str, str]]] = None
    """
    Информация о привязанных банковских картах.
    Если к счету не привязано ни одной карты, параметр отсутствует.
    Если к счету привязана хотя бы одна карта, параметр содержит список данных о привязанных картах.
    pan_fragment	string	Маскированный номер карты.
    type	string	
    Тип карты. Может отсутствовать, если неизвестен. Возможные значения:
    - VISA;
    - MasterCard;
    - AmericanExpress;
    - JCB.
    """


class OperationType(Enum):
    """
    Типы операций YooMoney

    deposition — пополнение счета (приход);
    payment — платежи со счета (расход);
    incoming_transfers_unaccepted —
    непринятые входящие P2P-переводы любого типа.

    """
    DEPOSITION = 'deposition'
    """Пополнение счета (приход);"""

    PAYMENT = 'payment'
    """Платежи со счета (расход)"""

    INCOMING = 'incoming-transfers-unaccepted'
    """непринятые входящие P2P-переводы любого типа."""


ALL_OPERATION_TYPES = [OperationType.DEPOSITION, OperationType.PAYMENT, OperationType.INCOMING]


@dataclass(frozen=True)
class Operation:
    operation_id: str
    """Идентификатор операции."""

    status: str
    """
    Статус платежа (перевода). Может принимать следующие значения:
    - success — платеж завершен успешно;
    - refused — платеж отвергнут получателем или отменен отправителем;
    - in_progress — платеж не завершен, перевод не принят получателем или ожидает ввода кода протекции.
    """

    operation_date: str
    """Дата и время совершения операции в формате строки в ISO формате с часовым поясом UTC."""

    title: str
    """Краткое описание операции (название магазина или источник пополнения)."""

    direction: str
    """
    Направление движения средств. Может принимать значения:
    - in (приход);
    - out (расход).
    """

    amount: Union[int, float]
    """Сумма операции."""

    operation_type: str
    """Тип операции. Возможные значения:
    payment-shop — исходящий платеж в магазин;
    outgoing-transfer — исходящий P2P-перевод любого типа;
    deposition — зачисление;
    incoming-transfer — входящий перевод или перевод до востребования;
    incoming-transfer-protected — входящий перевод с кодом протекции.
    """

    label: Optional[str] = None
    """Метка платежа.
     Присутствует для входящих и исходящих переводов другим пользователям ЮMoney,
     у которых был указан параметр label вызова send()
     """

    pattern_id: Optional[str] = None
    """Идентификатор шаблона, по которому совершен платеж. Присутствует только для платежей."""

    details: Optional[Any] = None


@dataclass(frozen=True)
class OperationDetails:
    operation_id: Optional[str] = None
    """Идентификатор операции. Можно получить при вызове метода history()"""

    status: Optional[str] = None
    """Статус платежа (перевода). Можно получить при вызове метода history()"""

    amount: Optional[float] = None
    """Сумма операции (сумма списания со счета)."""

    operation_date: Optional[str] = None
    """Дата и время совершения операции в формате строки в ISO формате с часовым поясом UTC."""

    operation_type: Optional[str] = None
    """Тип операции. Возможные значения:
    payment-shop — исходящий платеж в магазин;
    outgoing-transfer — исходящий P2P-перевод любого типа;
    deposition — зачисление;
    incoming-transfer — входящий перевод или перевод до востребования;
    incoming-transfer-protected — входящий перевод с кодом протекции.
    """

    direction: Optional[Literal['in', 'out']] = None
    """
    Направление движения средств. Может принимать значения:
    - in (приход);
    - out (расход).
    """

    comment: Optional[str] = None
    """Комментарий к переводу или пополнению. Присутствует в истории отправителя перевода или получателя пополнения."""

    digital_goods: Optional[Dict[str, Dict[str, List[Dict[str, str]]]]] = None
    """
    Данные о цифровом товаре (пин-коды и бонусы игр, iTunes, Xbox, etc.)
    Поле присутствует при успешном платеже в магазины цифровых товаров.
    Описание формата: https://yoomoney.ru/docs/wallet/process-payments/process-payment#digital-goods
    """

    details: Optional[str] = None
    """
    Детальное описание платежа.
    Строка произвольного формата, может содержать любые символы и переводы строк.
    Необязательный параметр.
    """

    label: Optional[str] = None
    """Метка платежа."""

    answer_datetime: Optional[str] = None
    """
    Дата и время приема или отмены перевода, защищенного кодом протекции.
    Присутствует для входящих и исходящих переводов, защищенных кодом протекции,
     если при вызове метода апи send вы указали protect=True в агрументах.
    Если перевод еще не принят или не отвергнут получателем, поле отсутствует.
    """

    expires: Optional[str] = None
    """
    Дата и время истечения срока действия кода протекции.
    Присутствует для входящих и исходящих переводов (от/другим) пользователям, защищенных кодом протекции,
    если при вызове метода апи send вы указали protect=True в агрументах.
    """

    protection_code: Optional[str] = None
    """Код протекции. Присутствует для исходящих переводов, защищенных кодом протекции."""

    codepro: Optional[bool] = None
    """Перевод защищен кодом протекции. Присутствует для переводов другим пользователям."""

    message: Optional[str] = None
    """Сообщение получателю перевода. Присутствует для переводов другим пользователям."""

    recipient_type: Optional[Literal['account', 'phone', 'email']] = None
    """
    Тип идентификатора получателя перевода. Возможные значения:
    account — номер счета получателя в сервисе ЮMoney;
    phone — номер привязанного мобильного телефона получателя;
    email — электронная почта получателя перевода.
    Присутствует для исходящих переводов другим пользователям.
    """

    recipient: Optional[str] = None
    """Идентификатор получателя перевода. Присутствует для исходящих переводов другим пользователям."""

    sender: Optional[str] = None
    """Номер счета отправителя перевода. Присутствует для входящих переводов от других пользователей."""

    title: Optional[str] = None
    """Краткое описание операции (название магазина или источник пополнения)."""

    fee: Optional[float] = None
    """Сумма комиссии. Присутствует для исходящих переводов другим пользователям."""

    amount_due: Optional[float] = None
    """Сумма к получению. Присутствует для исходящих переводов другим пользователям."""

    pattern_id: Optional[str] = None
    """Идентификатор шаблона платежа, по которому совершен платеж. Присутствует только для платежей."""

    error: Optional[str] = None
    """
    Код ошибки, присутствует при ошибке выполнения запрос
    Возможные ошибки:
    illegal_param_operation_id - неверное значение параметра operation_id
    Все прочие значения - техническая ошибка, повторите вызов метода позднее.
    """


@dataclass
class PreProcessPaymentResponse:
    """
    Объект, который вы получаете при вызове _pre_process_payment.
    При вызове данного метода вы не списываете деньги со своего счёта, а условно подготавливаете его к отправке.
    Для отправки денег на счёт используйте метод send()
    """
    status: Literal['success', 'refused']
    request_id: str
    recipient_account_status: Literal['anonymous', 'named', 'identified']
    fees: Dict[str, float]
    balance: Optional[float] = None
    recipient_account_type: Optional[Literal['personal', 'professional']] = None
    recipient_identified: bool = False
    recipient_masked_account: Optional[str] = None
    multiple_recipients_found: Optional[str] = None
    contract_amount: Optional[float] = None
    error: Optional[str] = None
    money_source: Optional[Dict[str, Dict[str, Union[str, bool, list]]]] = None
    protection_code: Optional[str] = None
    account_unblock_uri: Optional[str] = None
    ext_action_uri: Optional[str] = None


@dataclass
class Payment:
    status: Literal['success', 'refused', 'in_progress', 'ext_auth_required']
    """
    Код результата выполнения операции. Возможные значения:
    success — успешное выполнение (платеж проведен). Это конечное состояние платежа.
    refused — отказ в проведении платежа. Причина отказа возвращается в поле error. Это конечное состояние платежа.
    in_progress — авторизация платежа не завершена.
     Приложению следует повторить запрос с теми же параметрами спустя некоторое время.
    ext_auth_required — для завершения авторизации платежа с использованием банковской карты
     требуется аутентификация по технологии 3‑D Secure.
    все прочие значения — состояние платежа неизвестно. Приложению следует
     повторить запрос с теми же параметрами спустя некоторое время.
    """

    payment_id: str
    """Идентификатор проведенного платежа. Присутствует только при успешном выполнении метода send()."""

    credit_amount: Optional[float] = None
    """
    Сумма, поступившая на счет получателя.
    Присутствует при успешном переводе средств на счет другого пользователя ЮMoney.
    """

    payer: Optional[str] = None
    """
    Номер счета плательщика. Присутствует при успешном переводе средств на счет другого пользователя ЮMoney.
    """

    payee: Optional[str] = None
    """
    Номер счета получателя. Присутствует при успешном переводе средств на счет другого пользователя ЮMoney.
    """

    payee_uid: Union[str, int, None] = None
    """Служебное значение, не фигурирует в документации"""

    invoice_id: Optional[str] = None
    """Номер транзакции магазина в ЮMoney. Присутствует при успешном выполнении платежа в магазин."""

    balance: Optional[float] = None
    """
    Баланс счета пользователя после проведения платежа. Присутствует только при выполнении следующих условий:
    - метод выполнен успешно;
    - токен авторизации обладает правом account-info.
    """

    error: Optional[str] = None
    """
    Код ошибки при проведении платежа (пояснение к полю status). Присутствует только при ошибках.
    """
    account_unblock_uri: Optional[str] = None
    """
    Адрес, на который необходимо отправить пользователя для разблокировки счета.
    Поле присутствует в случае ошибки account_blocked.
    """

    acs_uri: Optional[str] = None

    acs_params: Optional[str] = None

    next_retry: Optional[int] = None
    """
    Рекомендуемое время, спустя которое следует повторить запрос, в миллисекундах.
     Поле присутствует при status=in_progress.
    """

    digital_goods: Optional[Dict[str, Dict[str, List[Dict[str, str]]]]] = None

    protection_code: Optional[str] = None
    """
    Код протекции, который был сгенерирован, если при вызове метода апи send вы указали protect=True в агрументах
    """


@dataclass(frozen=True)
class IncomingTransaction:
    status: Literal['success', 'refused']
    protection_code_attempts_available: int
    ext_action_uri: Optional[str] = None
    error: Optional[str] = None


__all__ = (
    'Response', 'Identification', 'WrapperData',
    'ProxyService', 'proxy_list', 'AccountInfo', 'OperationType', 'ALL_OPERATION_TYPES', 'Operation',
    'OperationDetails',
    'PreProcessPaymentResponse', 'Payment', 'IncomingTransaction'
)



from typing import Union, Optional, Any

from pydantic import BaseModel, Field

from glQiwiApi.utils import custom_load


class Sum(BaseModel):
    amount: Union[int, float, str]
    currency: str

    class Config:
        json_loads = custom_load


class OptionalSum(BaseModel):
    value: Union[int, float]
    currency: str


class Commission(BaseModel):
    provider_id: int = Field(alias="providerId")
    withdraw_sum: Sum = Field(alias="withdrawSum")
    enrollment_sum: Sum = Field(alias="enrollmentSum")
    qiwi_commission: Sum = Field("qwCommission")
    withdraw_to_enrollment_rate: int = Field(alias="withdrawToEnrollmentRate")

    class Config:
        json_loads = custom_load


class Type(BaseModel):
    id: str
    title: str


class BillMixin:
    _w = None
    bill_id: Optional[str] = None

    def initialize(self, w: Any):
        self._w = w
        return self

    async def check(self) -> bool:
        async with self._w:
            return (await self._w.check_p2p_bill_status(bill_id=self.bill_id)) == 'PAID'


__all__ = [
    'Sum', 'OptionalSum', 'Commission', 'Type', 'BillMixin'
]



from .account import Account
from .account_info import QiwiAccountInfo
from .bill import Bill, BillError
from .limit import Limit
from .stats import Statistic
from .transaction import Transaction

__all__ = [
    'QiwiAccountInfo', 'Transaction', 'Bill', 'BillError','Statistic', 'Limit', 'Account'
]



import time

from glQiwiApi.data import WrapperData

BASE_QIWI_URL = 'https://edge.qiwi.com'

DEFAULT_QIWI_HEADERS = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
    'Authorization': 'Bearer {token}',
    'Host': 'edge.qiwi_models.com',
}

ERROR_CODE_NUMBERS = {
    "400": "Ошибка синтаксиса запроса (неправильный формат данных)."
           "Если вы используете to_wallet или to_card, вероятно, у вас недостаточно"
           "денег для перевода или вы указали неверный номер счета для перевода",
    "401": "Неверный токен или истек срок действия токена API",
    "403": "Нет прав на данный запрос (недостаточно разрешений у токена API)",
    "404": "Не найдена транзакция или отсутствуют платежи с указанными признаками",
    "423": "Слишком много запросов, сервис временно недоступен",
    "405": "Ошибка, связанная с типом запроса к апи, обратитесь к разработчику или откройте issue",
    "500": "Внутренняя ошибка сервиса"
}

P2P_QIWI_HEADERS = {
    'Authorization': 'Bearer {token}',
    'Accept': 'application/json',
    'Content-Type': 'application/json',
}

QIWI_TO_CARD = WrapperData(
    json=
    {
        "id": str(int(time.time() * 1000)),
        "sum":
            {
                "amount": "", "currency": "643"
            },

        "paymentMethod":
            {
                "type": "Account", "accountId": "643"
            },
        "fields": {
            "account": ""
        }
    },
    headers=DEFAULT_QIWI_HEADERS
)

P2P_DATA = WrapperData(
    json={
        "amount": {
            "currency": "RUB",
            "value": "{amount}"
        },
        "expirationDateTime": "",
        "comment": "{comment}",
        "customFields": {
            "paySourcesFilter": "qw",
            "themeCode": "Yvan-YKaSh",
        }
    },

    headers=P2P_QIWI_HEADERS
)

QIWI_TO_WALLET = WrapperData(
    json={
        "id": str(int(time.time() * 1000)),
        "sum": {
            "amount": "",
            "currency": ""
        },
        "paymentMethod": {
            "type": "Account",
            "accountId": "643"
        },
        "comment": "",
        "fields": {
            "account": ""
        }
    },
    headers=DEFAULT_QIWI_HEADERS
)

TRANSACTION_TRANSFER = {
    'txnId': 'transaction_id',
    'personId': 'person_id',
    'account': 'to_account'
}

IDENTIFICATION_TRANSFER = {
    'id': 'identification_id',
    'firstName': 'first_name',
    'middleName': 'middle_name',
    'lastName': 'last_name',
    'birthDate': 'birth_date'
}

P2P_BILL_TRANSFER = {
    'siteId': 'site_id',
    'billId': 'bill_id',
    'creationDateTime': 'creation_date_time',
    'expirationDateTime': 'expiration_date_time',
    'payUrl': 'pay_url',
    'customFields': 'custom_fields',
    'value': 'amount'
}

LIMIT_TYPES = ['TURNOVER', 'REFILL', 'PAYMENTS_P2P', 'PAYMENTS_PROVIDER_INTERNATIONALS', 'PAYMENTS_PROVIDER_PAYOUT',
               'WITHDRAW_CASH']

LIMIT_TYPES_TRANSFER = {
    'max': 'max_limit',
    'type': 'limit_type'
}

ONLINE_COMMISSION_DATA = {
    "account": "",
    "paymentMethod": {
        "type": "Account",
        "accountId": "643"
    },
    "purchaseTotals":
        {
            "total":
                {"amount": "",
                 "currency": "643"
                 }
        }
}

COMMISSION_TRANSFER = {
    'providerId': 'provider_id',
    'withdrawSum': 'withdraw_sum',
    'qwCommission': 'qw_commission',
    'withdrawToEnrollmentRate': 'withdraw_to_enrollment_rate'
}
__all__ = (
    'LIMIT_TYPES', 'TRANSACTION_TRANSFER', 'IDENTIFICATION_TRANSFER', 'QIWI_TO_WALLET', 'QIWI_TO_CARD',
    'DEFAULT_QIWI_HEADERS', 'LIMIT_TYPES_TRANSFER', 'P2P_QIWI_HEADERS', 'P2P_DATA', 'P2P_BILL_TRANSFER',
    'ONLINE_COMMISSION_DATA', 'COMMISSION_TRANSFER', 'ERROR_CODE_NUMBERS', 'BASE_QIWI_URL'
)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    packages=setuptools.find_packages(),
    include_package_data=True,
    name="glQiwiApi",  # Replace with your own username
    version="0.2.1",
    author="GLEF1X",
    author_email="glebgar567@gmail.com",
    description="Parser for post and get requests",
    # Длинное описание, которое будет отображаться на странице PyPi. Использует README.md репозитория для заполнения.
    long_description=long_description,
    # Определяет тип контента, используемый в long_description.
    long_description_content_type="text/markdown",
    url="https://github.com/GLEF1X/glQiwiApi",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pytz==2021.1",
        "aiohttp==3.7.3",
        "aiosocksy==0.1.2",
        'aiofiles==0.6.0',
        "pydantic==1.8.1",
        "wheel",
    ],
    python_requires=">=3.7",
)



from typing import Optional

from pydantic import BaseModel, Field

from glQiwiApi.data_models.basics import Sum, Type
from glQiwiApi.utils import custom_load


class Account(BaseModel):
    alias: str
    fs_alias: str = Field(alias="fsAlias")
    bank_alias: str = Field(alias="bankAlias")
    title: str
    has_balance: bool = Field(alias="hasBalance")
    balance: Optional[Sum] = Field(const=None)
    currency: int
    account_type: Optional[Type] = Field(None, alias="type")
    is_default_account: bool = Field(alias="defaultAccount")

    class Config:
        json_loads = custom_load


__all__ = [
    'Account'
]



from typing import List

from pydantic import BaseModel, Field

from glQiwiApi.data_models.basics import Sum
from glQiwiApi.utils import custom_load


class Statistic(BaseModel):
    incoming: List[Sum] = Field(alias="incomingTotal")
    out: List[Sum] = Field(alias="outgoingTotal")

    class Config:
        json_loads = custom_load


__all__ = [
    'Statistic'
]



from datetime import datetime
from functools import lru_cache
from typing import List, Dict, Any, Union, Optional, Tuple, Literal

from glQiwiApi.api import HttpXParser
from glQiwiApi.abstracts import AbstractPaymentWrapper
from glQiwiApi.aiohttp_custom_api import CustomParser
from glQiwiApi.data import AccountInfo, OperationType, Operation, OperationDetails, PreProcessPaymentResponse, \
    Payment, IncomingTransaction
from glQiwiApi.exceptions import NoUrlFound, InvalidData
from glQiwiApi.utils import parse_auth_link, parse_headers, DataFormatter, datetime_to_str_in_iso
from glQiwiApi.yoo_money.basic_yoomoney_config import *


class YooMoneyAPI(AbstractPaymentWrapper):
    """
    Класс, реализующий обработку запросов к YooMoney, используя основной класс HttpXParser,
    удобен он тем, что не просто отдает json подобные объекты, а всё это конвертирует в python датаклассы.
    Для работы с данным классом, необходимо загестрировать токен по такому
    """

    def __init__(self, api_access_token: str, without_context: bool = False) -> None:
        """
        Конструктор принимает токен, олученный из класс метода get_access_token() этого же класса
        и специальный аттрибут without_context,

        :param api_access_token: апи токен для запросов
        :param without_context: булевая переменная, указывающая будет ли объект класса "глобальной" переменной
         или будет использована в async with контексте
        """
        self.api_access_token = api_access_token
        self._parser = CustomParser(without_context=without_context, messages=ERROR_CODE_NUMBERS)
        self._formatter = DataFormatter()

    async def __aenter__(self) -> 'YooMoneyAPI':
        """Создаем сессию, чтобы не пересоздавать её много раз"""
        self._parser.create_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Закрываем сессию при выходе"""
        await self._parser.session.close()

    def _auth_token(self, headers: dict) -> Dict[Any, Any]:
        headers['Authorization'] = headers['Authorization'].format(
            token=self.api_access_token
        )
        return headers

    @classmethod
    async def build_url_for_auth(cls, scope: List[str], client_id: str,
                                 redirect_uri: str = 'https://example.com') -> str:
        """
        Метод для получения ссылки для дальнейшей авторизации и получения токена yoomoney\n
        Сценарий авторизации, взятый из документации, приложения пользователем:\n
        1. Пользователь инициирует авторизацию приложения для управления своим счетом.\n
        2. Приложение отправляет запрос Authorization Request на сервер ЮMoney.\n
        3. ЮMoney перенаправляют пользователя на страницу авторизации.\n
        4. Пользователь вводит свой логин и пароль,
           просматривает список запрашиваемых прав и подтверждает, либо отклоняет запрос авторизации.
        5. Приложение получает ответ Authorization Response
         в виде HTTP Redirect со временным токеном для получения доступа или кодом ошибки.\n
        6. Приложение, используя полученный временный токен доступа,
           отправляет запрос на получение токена авторизации (Access Token Request).\n
        7. Ответ содержит токен авторизации (access_token).\n
        8. Приложение сообщает пользователю результат авторизации.\n
        Подробно это описано в оффициальной документации:
        https://yoomoney.ru/docs/wallet/using-api/authorization/request-access-token

        :param scope: OAuth2-авторизации приложения пользователем, права перечисляются через пробел.
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда прийдет временный код, который нужен для получения основного токена
        :return: ссылку, по которой нужно перейти и сделать авторизацию через логин/пароль
        """
        headers = parse_headers()
        params = {
            'client_id': client_id,
            'response_type': 'code',
            'redirect_uri': redirect_uri,
            'scope': " ".join(scope)
        }
        async for response in HttpXParser().fast().fetch(
                url=BASE_YOOMONEY_URL + '/oauth/authorize',
                headers=headers,
                data=params,
                method='POST'
        ):
            try:
                return parse_auth_link(response.response_data)
            except IndexError:
                raise NoUrlFound('Не удалось найти ссылку для авторизации в ответе от апи, проверятьте client_id')

    @classmethod
    async def get_access_token(cls, code: str, client_id: str, redirect_uri: str = 'https://example.com') -> str:
        """
        Метод для получения токена для запросов к YooMoney API

        :param code: временный код, который был получен в методе base_authorize()
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда прийдет временный код, который нужен для получения основного токена
        :return: YooMoney API TOKEN
        """
        headers = parse_headers(content_json=True)
        params = {
            'code': code,
            'client_id': client_id,
            'grant_type': 'authorization_code',
            'redirect_uri': redirect_uri
        }
        async for response in HttpXParser().fast().fetch(
                url=BASE_YOOMONEY_URL + '/oauth/token',
                headers=headers,
                data=params,
                method='POST',
                get_json=True
        ):
            return response.response_data.get('access_token')

    async def revoke_api_token(self) -> Optional[Dict[str, bool]]:
        """
        Метод для отзывания токена, при этом все его права тоже пропадают
        Документация: https://yoomoney.ru/docs/wallet/using-api/authorization/revoke-access-token
        """
        headers = self._auth_token(parse_headers(auth=True))
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/revoke',
                method='POST',
                headers=headers
        ):
            if response.ok:
                return {'success': True}

    @lru_cache
    async def account_info(self) -> AccountInfo:
        """
        Метод для получения информации об аккаунте пользователя
        Подробная документация: https://yoomoney.ru/docs/wallet/user-account/account-info

        :return: объект AccountInfo
        """
        headers = self._auth_token(parse_headers(**content_and_auth))
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/account-info',
                headers=headers,
                method='POST'
        ):
            try:
                return self._formatter.format_objects(
                    iterable_obj=(response.response_data,),
                    obj=AccountInfo,
                )[0]
            except IndexError:
                raise InvalidData('Cannot fetch account info, check your token')

    async def transactions(
            self,
            operation_types: Optional[Union[List[OperationType], Tuple[OperationType]]] = None,
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None,
            start_record: Optional[int] = None,
            records: int = 30,
            label: Optional[Union[str, int]] = None
    ) -> List[Operation]:
        """
        Подробная документация: https://yoomoney.ru/docs/wallet/user-account/operation-history\n
        Метод позволяет просматривать историю операций (полностью или частично) в постраничном режиме.\n
        Записи истории выдаются в обратном хронологическом порядке: от последних к более ранним.\n
        Перечень типов операций, которые требуется отобразить. Возможные значения:\n
        DEPOSITION — пополнение счета (приход);\n
        PAYMENT — платежи со счета (расход);\n
        INCOMING(incoming-transfers-unaccepted) — непринятые входящие P2P-переводы любого типа.\n

        :param operation_types: Тип операций
        :param label: string.
        Отбор платежей по значению метки.
         Выбираются платежи, у которых указано заданное значение параметра label вызова request-payment.
        :param start_date: datetime	Вывести операции от момента времени (операции, равные start_date, или более поздние)
        Если параметр отсутствует, выводятся все операции.
        :param end_date: datetime	Вывести операции до момента времени (операции более ранние, чем end_date).
         Если параметр отсутствует, выводятся все операции.
        :param start_record: string
         Если параметр присутствует, то будут отображены операции, начиная с номера start_record.
         Операции нумеруются с 0. Подробнее про постраничный вывод списка
        :param records:	int	Количество запрашиваемых записей истории операций.
         Допустимые значения: от 1 до 100, по умолчанию — 30.
        """
        headers = self._auth_token(parse_headers(**content_and_auth))

        if records <= 0 or records > 100:
            raise InvalidData(
                'Неверное количество записей. '
                'Кол-во записей, которые можно запросить, находиться в диапазоне от 1 до 100 включительно'
            )

        data = {
            'records': records
        }

        if operation_types:
            if all(isinstance(operation_type, OperationType) for operation_type in operation_types):
                data.update({'type': ' '.join([operation_type.value for operation_type in operation_types])})

        if isinstance(start_record, int):
            if start_record < 0:
                raise InvalidData('Укажите позитивное число')
            data.update({'start_record': start_record})

        data.update({'label': label}) if isinstance(label, str) else None
        if start_date:
            if not isinstance(start_date, datetime):
                raise InvalidData('Параметр start_date был передан неправильным типом данных')
            data.update({'from': datetime_to_str_in_iso(start_date, yoo_money_format=True)})

        if end_date:
            if not isinstance(end_date, datetime):
                raise InvalidData('Параметр end_date был передан неправильным типом данных')
            data.update({'till': datetime_to_str_in_iso(end_date, yoo_money_format=True)})

        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/operation-history',
                method='POST',
                headers=headers,
                data=data,
                get_json=True
        ):
            return self._formatter.format_objects(
                iterable_obj=response.response_data.get('operations'),
                obj=Operation,
                transfers=OPERATION_TRANSFER
            )

    async def transaction_info(self, operation_id: str) -> OperationDetails:
        """
        Позволяет получить детальную информацию об операции из истории.
        Требуемые права токена: operation-details.
        Более подробная документация: https://yoomoney.ru/docs/wallet/user-account/operation-details

        :param operation_id: Идентификатор операции
        """
        headers = self._auth_token(parse_headers(**content_and_auth))
        payload = {
            'operation_id': operation_id
        }
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/operation-details',
                headers=headers,
                data=payload
        ):
            try:
                obj = self._formatter.format_objects(
                    iterable_obj=(response.response_data,),
                    obj=OperationDetails,
                    transfers=OPERATION_TRANSFER
                )[0]
                if obj.error:
                    raise IndexError()
                return obj
            except (IndexError, AttributeError):
                raise ConnectionError(
                    'Не удалось получить объект, проверьте токен, который вы передаете в конструктор класса'
                )

    async def _pre_process_payment(
            self,
            to_account: str,
            amount: Union[int, float],
            pattern_id: str = 'p2p',
            comment_for_history: Optional[str] = None,
            comment_for_receiver: Optional[str] = None,
            protect: bool = False,
            expire_period: int = 1
    ) -> PreProcessPaymentResponse:
        """
        Более подробная документация: https://yoomoney.ru/docs/wallet/process-payments/request-payment\n
        Создание платежа, проверка параметров и возможности приема платежа магазином или перевода средств
        на счет пользователя ЮMoney.\n
        Данный метод не рекомендуется использовать напрямую, гораздо проще использовать send.
        Требуемые права токена:
        payment.to-account («идентификатор получателя», «тип идентификатора») или payment-p2p.

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода (номер счета, номер телефона или email).
        :param amount: Сумма к получению (придет на счет получателя счет после оплаты).
        :param comment_for_history: Комментарий к переводу, отображается в истории отправителя.
        :param comment_for_receiver:	string	Комментарий к переводу, отображается получателю.
        :param protect: Значение параметра true — признак того, что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции и получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365. Необязательный параметр. По умолчанию 1.
        """
        headers = self._auth_token(parse_headers(**content_and_auth))
        payload = {
            'pattern_id': pattern_id,
            'to': to_account,
            'amount_due': amount,
            'comment': comment_for_history,
            'message': comment_for_receiver,
            'expire_period': expire_period,
            'codepro': protect
        }
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/request-payment',
                headers=headers,
                data=payload
        ):
            try:
                obj = self._formatter.format_objects(
                    iterable_obj=(response.response_data,),
                    obj=PreProcessPaymentResponse
                )[0]
                if obj.status != 'success':
                    raise IndexError()
                return obj
            except (IndexError, AttributeError):
                raise ConnectionError(
                    'Не удалось создать pre_payment запрос, проверьте переданные параметры и попробуйте ещё раз'
                )

    async def send(
            self,
            to_account: str,
            amount: Union[int, float],
            money_source: Literal['wallet', 'card'] = 'wallet',
            pattern_id: str = 'p2p',
            cvv2_code: str = '',
            card_type: Optional[Literal['Visa', 'MasterCard', 'American Express', 'JCB']] = None,
            protect: bool = False,
            comment_for_history: Optional[str] = None,
            comment: Optional[str] = None,
            expire_period: int = 1
    ) -> Payment:
        """
        Метод для отправки денег на аккаунт или карту другого человека. Данная функция делает сразу 2 запроса,
        из-за этого вы можете почувствовать небольшую потерю производительности, вы можете использовать метод
        _pre_process_payment и получать объект PreProcessPaymentResponse,
        в котором есть информация о ещё неподтвержденном платеже\n
        Более подробная документация: https://yoomoney.ru/docs/wallet/process-payments/process-payment

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода (номер счета, номер телефона или email).
        :param amount: Сумма к получению (придет на счет получателя счет после оплаты). МИНИМАЛЬНАЯ СУММА 2.
        :param money_source: Запрашиваемый метод проведения платежа.
         wallet — со счета пользователя, если вы хотите использовать card,
         тогда нужно будет передать card_type для поиска карты в списке ваших банковских карт,
         а также опционально cvv2 код для проведения платежа
        :param comment_for_history: Комментарий к переводу, отображается в истории отправителя.
        :param card_type: Тип банковской карты, нужно заполнять, только если вы хотите списать средства с вашей карты
        :param cvv2_code: опционально, может быть не передан, однако, если для оплаты картой это требуется,
         параметр стоит передавать
        :param comment:	Комментарий к переводу, отображается получателю.
        :param protect: Значение параметра true — признак того, что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции и получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365. Необязательный параметр. По умолчанию 1.
        """
        if amount < 2:
            raise InvalidData('Введите сумму, которая больше минимальной(2 и выше)')
        pre_payment = await self._pre_process_payment(
            to_account=to_account,
            amount=amount,
            comment_for_history=comment_for_history,
            comment_for_receiver=comment,
            expire_period=expire_period,
            protect=protect,
            pattern_id=pattern_id
        )
        headers = self._auth_token(parse_headers(**content_and_auth))

        payload = {
            'request_id': pre_payment.request_id,
            'money_source': 'wallet'
        }

        if money_source == 'card' and isinstance(pre_payment, PreProcessPaymentResponse):
            if pre_payment.money_source.get('cards').get('allowed') == 'true':
                if not card_type:
                    payload.update({
                        'money_source': pre_payment.money_source.get('cards').get('items')[0].get('id'),
                        'csc': cvv2_code
                    })
                else:
                    cards = pre_payment.money_source.get('cards').get('items')
                    for card in cards:
                        if card.get('type') == card_type:
                            payload.update(
                                {'money_source': card.get('id'), 'csc': cvv2_code},
                            )
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/process-payment',
                method='POST',
                headers=headers,
                data=payload
        ):
            try:
                obj = self._formatter.format_objects(
                    iterable_obj=(response.response_data,),
                    obj=Payment
                )[0]
                obj.protection_code = pre_payment.protection_code
                return obj
            except (IndexError, AttributeError):
                raise ConnectionError(
                    'Не удалось создать запрос на перевод средств, проверьте переданные параметры и попробуйте ещё раз'
                )

    async def get_balance(self) -> Optional[Union[float, int]]:
        """Метод для получения баланса на кошельке юмани"""
        return (await self.account_info()).balance

    async def accept_incoming_transaction(
            self,
            operation_id: str,
            protection_code: str
    ) -> IncomingTransaction:
        """
        Прием входящих переводов, защищенных кодом протекции, если вы передали в метод send параметр protect,
        и переводов до востребования.
        Количество попыток приема входящего перевода с кодом протекции ограничено.
        При исчерпании количества попыток, перевод автоматически отвергается (перевод возвращается отправителю).
        Более подробная документация: https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-accept

        :param operation_id: Идентификатор операции, значение параметра operation_id ответа метода history()
        :param protection_code: Код протекции. Строка из 4-х десятичных цифр.
         Указывается для входящего перевода, защищенного кодом протекции. Для переводов до востребования отсутствует.
        """
        headers = self._auth_token(parse_headers(**content_and_auth))
        payload = {
            'operation_id': operation_id,
            'protection_code': protection_code
        }
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/incoming-transfer-accept',
                headers=headers,
                data=payload,
                method='POST',
                get_json=True
        ):
            try:
                return self._formatter.format_objects(
                    iterable_obj=(response.response_data,),
                    obj=IncomingTransaction
                )[0]
            except IndexError:
                raise ConnectionError('Такая транзакция не найдена или были переданы невалид данные')

    async def reject_transaction(self, operation_id: str) -> Dict[str, str]:
        """
        Отмена входящих переводов, защищенных кодом протекции, если вы передали в метод send параметр protect,
        и переводов до востребования. \n
        При отмене перевода он возвращается отправителю. \n
        Требуемые права токена: incoming-transfers

        Более подробная документация: https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-reject

        :param operation_id: Идентификатор операции, значение параметра operation_id ответа метода history().
        :return: словарь в json формате с ответом от апи
        """
        headers = self._auth_token(parse_headers(**content_and_auth))
        async for response in self._parser.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/incoming-transfer-reject',
                headers=headers,
                data={'operation_id': operation_id},
                method='POST',
                get_json=True
        ):
            return response.response_data

    async def check_transaction(
            self,
            amount: Union[int, float],
            transaction_type: Literal['in', 'out'] = 'in',
            comment: Optional[str] = None,
            rows_num: int = 100,
            sender_number: Optional[str] = None
    ) -> bool:
        """
        Метод для проверки транзакции.\n Рекомендуется использовать только если вы не можете написать свой обработчик.\n
        Данный метод использует self.transactions(rows_num=rows_num) для получения платежей.\n
        Для небольшой оптимизации вы можете уменьшить rows_num задав его, однако это не гарантирует правильный результат

        :param amount: сумма платежа
        :param transaction_type: тип платежа
        :param sender_number: номер получателя
        :param rows_num: кол-во платежей, которое будет проверяться
        :param comment: комментарий, по которому будет проверяться транзакция
        :return: bool, есть ли такая транзакция в истории платежей
        """
        transactions = await self.transactions(
            operation_types=[OperationType.DEPOSITION if transaction_type == 'in' else OperationType.PAYMENT],
            records=rows_num
        )

        for transaction in transactions:
            details_transaction = await self.transaction_info(transaction.operation_id)
            detail_amount = details_transaction.amount if transaction_type == 'in' else details_transaction.amount_due
            detail_comment = details_transaction.comment if transaction_type == 'in' else details_transaction.message
            if detail_amount >= amount and details_transaction.direction == transaction_type:
                if transaction.status == 'success':
                    if detail_comment == comment and details_transaction.sender == sender_number:
                        return True
                    elif isinstance(comment, str) and isinstance(sender_number, str):
                        continue
                    elif detail_comment == comment:
                        return True

        return False



# Ссылка для авторизации юмани
BASE_YOOMONEY_URL = 'https://yoomoney.ru'

DEFAULT_YOOMONEY_HEADERS = {
    'Host': 'yoomoney.ru',
    'Content-Type': 'application/x-www-form-urlencoded'
}

ERROR_CODE_NUMBERS = {
    '400': 'Ошибка, связанная с типом запроса к апи, возможно вы передали недействительный API токен.',
    '401': 'Указан несуществующий, просроченный, или отозванный токен.',
    '403': 'Запрошена операция, на которую у токена нет прав.'
}

OPERATION_TRANSFER = {
    'datetime': 'operation_date',
    'type': 'operation_type'
}

content_and_auth = {'content_json': True, 'auth': True}

__all__ = (
    'DEFAULT_YOOMONEY_HEADERS', 'ERROR_CODE_NUMBERS',
    'BASE_YOOMONEY_URL', 'OPERATION_TRANSFER', 'content_and_auth'
)
